@@ -55,10 +55,17 @@ class GlobalEpistasisModel:
 
     alphabet_dict: (str)
         Specifies the type of input sequences. Three possible choices
-        allowed: ['dna','rna','protein']
+        allowed: ['dna','rna','protein'].
 
-    sub_network_layers_nodes_dict: (dict)
-        Dictionary that specifies the number of layers and nodes in each subnetwork layer
+    custom_architecture: (tf.model)
+        a custom neural network architecture that replaces the
+        default architecture implemented.
+
+    ohe_single_batch_size: (int)
+        integer specifying how many sequences to one-hot encode at a time.
+        The larger this number number, the quicker the encoding will happen,
+        but this may also take up a lot of memory and throw an exception
+        if its too large. Currently for additive models only.
 
     """
 
@@ -68,14 +75,16 @@ class GlobalEpistasisModel:
                  model_type='additive',
                  test_size=0.2,
                  alphabet_dict='dna',
-                 sub_network_layers_nodes_dict=None):
+                 custom_architecture=None,
+                 ohe_single_batch_size=10000):
 
         # set class attributes
         self.X, self.y = X, y
         self.model_type = model_type
         self.test_size = test_size
         self.alphabet_dict = alphabet_dict
-        self.sub_network_layers_nodes_dict = sub_network_layers_nodes_dict
+        self.custom_architecture = custom_architecture
+        self.ohe_single_batch_size = ohe_single_batch_size
 
         # class attributes that are not parameters
         # but are useful for using trained models
@@ -113,7 +122,7 @@ class GlobalEpistasisModel:
 
         if model_type == 'additive':
             # one-hot encode sequences in batches in a vectorized way
-            self.input_seqs_ohe = onehot_encode_array(self.x_train, self.bases)
+            self.input_seqs_ohe = onehot_encode_array(self.x_train, self.bases, self.ohe_single_batch_size)
 
         elif model_type == 'neighbor':
             # one-hot encode sequences in batches in a vectorized way
@@ -140,7 +149,7 @@ class GlobalEpistasisModel:
         check(isinstance(self.y, (list, np.ndarray)),
               'type(y) = %s must be of type list or np.array' % type(self.y))
 
-        # check that alphabet_dict is valid
+        # check that model type valid
         check(self.model_type in {'additive', 'neighbor', 'pairwise'},
               'model_type = %s; must be "additive", "neighbor", or "pairwise"' %
               self.model_type)
@@ -148,28 +157,25 @@ class GlobalEpistasisModel:
     # JBK: User should be able to set these parameters in the constructor
     def define_model(self,
                      monotonic=True,
-                     regularization=None,
-                     activations='sigmoid'):
+                     custom_architecture=None):
 
         """
         Method that will define the architecture of the global epistasis model.
         using the tensorflow.keras functional api. If the subnetwork architecture
-        is not None, than archicture is constructed via the dict sub_network_layers_nodes_dict
+        is not None, than archicture is constructed via the self.custom_architecture
+        keyword
 
         parameters
         ----------
 
         monotonic: (boolean)
-            If True, than weights in subnetwork will be constrained to be non-negative.
-            Doing so will constrain the global epistasis non-linearity to be monotonic
+            Indicates whether to use monotonicity constraint in GE nonlinear function.
+            If true then weights of GE nonlinear function will be constraned to
+            be non-negative.
 
-        regularization: (str)
-            String that specifies type of regularization to use. Valid choices include
-            ['l2','dropout', ... link to tf docs].
-
-        activations: (str)
-            Activation function used in the non-linear sub-network. Link to allowed
-            activation functions from TF docs...
+        custom_architecture: (tf.model)
+            a custom neural network architecture that replaces the
+            default architecture implemented.
 
         returns
         -------
@@ -180,19 +186,44 @@ class GlobalEpistasisModel:
 
         """
 
-        number_input_layer_nodes = len(self.input_seqs_ohe[0])
-        inputTensor = Input((number_input_layer_nodes,), name='Sequence')
+        # user has not provided custom architecture, implement a default architecture
+        if custom_architecture is None:
 
-        phi = Dense(1, use_bias=True)(inputTensor)
+            number_input_layer_nodes = len(self.input_seqs_ohe[0])
+            inputTensor = Input((number_input_layer_nodes,), name='Sequence')
 
-        intermediateTensor = Dense(50, activation='sigmoid', kernel_constraint=nonneg())(phi)
-        outputTensor = Dense(1, kernel_constraint=nonneg())(intermediateTensor)
+            phi = Dense(1, use_bias=True)(inputTensor)
 
-        # create the model:
+            # implement monotonicity constraints
+            if monotonic:
+                intermediateTensor = Dense(50, activation='sigmoid', kernel_constraint=nonneg())(phi)
+                outputTensor = Dense(1, kernel_constraint=nonneg())(intermediateTensor)
+            else:
+                intermediateTensor = Dense(50, activation='sigmoid')(phi)
+                outputTensor = Dense(1)(intermediateTensor)
 
-        model = Model(inputTensor, outputTensor)
-        self.model = model
-        return model
+            # create the model:
+
+            model = Model(inputTensor, outputTensor)
+            self.model = model
+
+            return model
+
+        # if user has provided custom architecture
+        else:
+            self.model = custom_architecture
+            return custom_architecture
+
+    def return_model(self):
+        """
+
+        returns (GE model)
+            Helper method that returns the model attribute,
+            so that it may easily accessible from the mavenn
+            Model class
+        """
+
+        return self.model
 
     def compile_model(self,
                       optimizer='Adam',
@@ -435,7 +466,7 @@ class GlobalEpistasisModel:
             return ge_nonlinearity, input_range, latent_trait
         else:
             ge_nonlinearity = ge_model.predict(input_range)
-            return ge_nonlinearity, input_range
+            return ge_nonlinearity
 
 
 class NoiseAgnosticModel:
@@ -466,8 +497,15 @@ class NoiseAgnosticModel:
         Specifies the type of input sequences. Three possible choices
         allowed: ['dna','rna','protein']
 
-    noise_model_layers_nodes_dict: (dict)
-        Dictionary that specifies the number of layers and nodes in each noise model layer
+    custom_architecture: (tf.model)
+        a custom neural network architecture that replaces the
+        default architecture implemented.
+
+    ohe_single_batch_size: (int)
+        integer specifying how many sequences to one-hot encode at a time.
+        The larger this number number, the quicker the encoding will happen,
+        but this may also take up a lot of memory and throw an exception
+        if its too large. Currently for additive models only.
 
     """
 
@@ -477,7 +515,8 @@ class NoiseAgnosticModel:
                  model_type='additive',
                  test_size=0.2,
                  alphabet_dict='dna',
-                 noise_model_layers_nodes_dict=None):
+                 custom_architecture=None,
+                 ohe_single_batch_size=10000):
 
         # set class attributes
         self.X = X
@@ -485,7 +524,8 @@ class NoiseAgnosticModel:
         self.model_type = model_type
         self.test_size = test_size
         self.alphabet_dict = alphabet_dict
-        self.noise_model_layers_nodes_dict = noise_model_layers_nodes_dict
+        self.custom_architecture = custom_architecture
+        self.ohe_single_batch_size = ohe_single_batch_size
 
         # class attributes that are not parameters
         # but are useful for using trained models
@@ -523,7 +563,7 @@ class NoiseAgnosticModel:
 
         if model_type=='additive':
             # one-hot encode sequences in batches in a vectorized way
-            self.input_seqs_ohe = onehot_encode_array(self.x_train, self.bases)
+            self.input_seqs_ohe = onehot_encode_array(self.x_train, self.bases, self.ohe_single_batch_size)
 
         elif model_type=='neighbor':
             # one-hot encode sequences in batches in a vectorized way
@@ -544,9 +584,8 @@ class NoiseAgnosticModel:
         pass
 
     def define_model(self,
-                     nonneg_weights=True,
-                     regularization=None,
-                     activations='sigmoid'):
+                     monotonic=True,
+                     custom_architecture=None):
 
         """
         Method that will define the architecture of the global epistasis model.
@@ -556,17 +595,12 @@ class NoiseAgnosticModel:
         parameters
         ----------
 
-        nonneg_weights: (boolean)
-            If True, than weights in output will be constrained to be non-negative.
-
-        regularization: (str)
-            String that specifies type of regularization to use. Valid choices include
-            ['l2','dropout', ... link to tf docs].
-
-        activations: (str)
-            Activation function used in the non-linear sub-network. Link to allowed
-            activation functions from TF docs...
+        monotonic: (boolean)
+            If True, than weights in noise model will be constrained to be non-negative.
 
+        custom_architecture: (tf.model)
+            a custom neural network architecture that replaces the
+            default architecture implemented.
 
         returns
         -------
@@ -577,19 +611,29 @@ class NoiseAgnosticModel:
 
         """
 
-        number_input_layer_nodes = len(self.input_seqs_ohe[0])
-        inputTensor = Input((number_input_layer_nodes,), name='Sequence')
+        if custom_architecture is None:
+            number_input_layer_nodes = len(self.input_seqs_ohe[0])
+            inputTensor = Input((number_input_layer_nodes,), name='Sequence')
 
-        phi = Dense(1, use_bias=True, name='additive_weights')(inputTensor)
+            phi = Dense(1, use_bias=True, name='additive_weights')(inputTensor)
 
-        intermediateTensor = Dense(10, activation='sigmoid', kernel_constraint=nonneg())(phi)
-        outputTensor = Dense(np.shape(self.y_train[0])[0], activation='softmax', kernel_constraint=nonneg())(
-            intermediateTensor)
+            # implement monotonicity constraints
+            if monotonic:
+                intermediateTensor = Dense(10, activation='sigmoid', kernel_constraint=nonneg())(phi)
+                outputTensor = Dense(np.shape(self.y_train[0])[0], activation='softmax', kernel_constraint=nonneg())(
+                    intermediateTensor)
+            else:
+                intermediateTensor = Dense(10, activation='sigmoid')(phi)
+                outputTensor = Dense(np.shape(self.y_train[0])[0], activation='softmax')(intermediateTensor)
 
-        # #create the model:
-        model = Model(inputTensor, outputTensor)
-        self.model = model
-        return model
+
+            # #create the model:
+            model = Model(inputTensor, outputTensor)
+            self.model = model
+            return model
+        else:
+            self.model = custom_architecture
+            return custom_architecture
 
     def compile_model(self,
                       lr=0.005):
@@ -624,12 +668,11 @@ class NoiseAgnosticModel:
                            metrics=['categorical_accuracy'])
 
     def fit(self,
-                  validation_split=0.2,
-                  epochs=50,
-                  verbose=1,
-                  use_early_stopping=True,
-                  early_stopping_patience=50
-                  ):
+            validation_split=0.2,
+            epochs=50,
+            verbose=1,
+            use_early_stopping=True,
+            early_stopping_patience=50):
 
         """
 
@@ -686,8 +729,6 @@ class NoiseAgnosticModel:
         self.history = history
         return history
 
-
-
     def model_evaluate(self,
                        model,
                        data):
@@ -711,6 +752,30 @@ class NoiseAgnosticModel:
 
         pass
 
+    def return_loss(self):
+
+        """
+        Method that returns loss values.
+
+        returns
+        -------
+        history: (object)
+            self.attribute/object that contains loss history
+
+        """
+
+        return self.history
+
+    def return_model(self):
+        """
+
+        returns (GE model)
+            Helper method that returns the model attribute,
+            so that it may easily accessible from the mavenn
+            Model class
+        """
+
+        return self.model
 
     def noise_model(self,
                     sequences=None,
@@ -786,4 +851,4 @@ class NoiseAgnosticModel:
             return noise_function, input_range, latent_trait
         else:
             noise_function = noise_model.predict(input_range)
-            return noise_function, input_range
+            return noise_function
